/*
 * Copyright 2020 University of Liège
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "fBuilder.h"
#include "fProblem.h"
#include "fField.h"
#include "fBody.h"
#include "fWake.h"

#include "wTag.h"
#include "wElement.h"
#include "wNode.h"
#include "wMem.h"

#define PI 3.14159

using namespace tbox;
using namespace fpm;

/**
 * @brief Initialize the solver and perform sanity checks
 * @authors Adrien Crovato & Axel Dechamps
 */
Builder::Builder(std::shared_ptr<Problem> _pbl) : pbl(_pbl), valid(false)
{
    // Check the problem
    pbl->check();

    // Setup AIC matrices
    pbl->field ? nF = pbl->field->tag->elems.size() : nF = 0;
    nP = 0;
    for (auto body : pbl->bodies)
        nP += body->tag->elems.size();
    A = Eigen::MatrixXd::Zero(nP, nP);
    B = Eigen::MatrixXd::Zero(nP, nP);
    C = Eigen::MatrixXd::Zero(nP, nF);
    Cx = Eigen::MatrixXd::Zero(nP, nF);
    Cy = Eigen::MatrixXd::Zero(nP, nF);
    Cz = Eigen::MatrixXd::Zero(nP, nF);

    // global->local id
    int i = 0;
    for (auto body : pbl->bodies)
        for (auto e : body->tag->elems)
        {
            rows[e] = i;
            i++;
        }
}

/**
 * @brief Fill AIC matrices
 */
void Builder::run()
{
    // Update element memory
    pbl->updateMem();

    // body to body - body panels
    for (auto jbody : pbl->bodies)
    {
        for (auto ej : jbody->tag->elems)
        {
            // Initialisation of storage variables for the iteration
            Eigen::Matrix3d glob2loc;
            size_t nVertices = ej->nodes.size();
            Eigen::MatrixXd trsfCoordSrc(3, nVertices);

            // Computations related to source panels
            glob2loc = Glob2loc(false, ej); // Transformation matrix
            trsfCoordSrc = TrsfCoordSrc(false, ej, glob2loc); // Local coordinates of the source panel

            // body panels
            for (auto ibody : pbl->bodies)
            {
                for (auto ei : ibody->tag->elems)
                {
                    // Initialisation of storage variables for the iteration
                    Eigen::Vector3d trsfCoordTgt;
                    Eigen::Vector3d distance;
                    Eigen::Vector2d mutau;

                    // Computations related to target panels
                    trsfCoordTgt = TrsfCoordTgt(false, ei, ej, glob2loc); // Local coordinates of the target panel
                    distance = Distance(false, ei, ej, glob2loc); // Distances between source and target panels
                    mutau = MuTau(false, ei, ej, trsfCoordSrc, trsfCoordTgt, distance); // Surface AIC

                    // Assembly of the matrices
                    A(rows.at(ei), rows.at(ej)) = mutau(0);
                    B(rows.at(ei), rows.at(ej)) = mutau(1);
                }
            }
        }
    }

    // body to body - body panels - symmetry
    if (pbl->symY)
    {
        for (auto jbody : pbl->bodies)
        {
            for (auto ej : jbody->tag->elems)
            {
                // Initialisation of storage variables for the iteration
                Eigen::Matrix3d glob2loc;
                size_t nVertices = ej->nodes.size();
                Eigen::MatrixXd trsfCoordSrc(3, nVertices);

                // Computations related to source panels
                glob2loc = Glob2loc(true, ej); // Transformation matrix
                trsfCoordSrc = TrsfCoordSrc(true, ej, glob2loc); // Local coordinates of the source panel

                // body panels
                for (auto ibody : pbl->bodies)
                {
                    for (auto ei : ibody->tag->elems)
                    {
                        // Initialisation of storage variables for the iteration
                        Eigen::Vector3d trsfCoordTgt;
                        Eigen::Vector3d distance;
                        Eigen::Vector2d mutau;

                        // Computations related to target panels
                        trsfCoordTgt = TrsfCoordTgt(true, ei, ej, glob2loc); // Local coordinates of the target panel
                        distance = Distance(true, ei, ej, glob2loc); // Distances between source and target panels
                        mutau = MuTau(true, ei, ej, trsfCoordSrc, trsfCoordTgt, distance); // Surface AIC

                        // Assembly of the matrices
                        A(rows.at(ei), rows.at(ej)) -= mutau(0);
                        B(rows.at(ei), rows.at(ej)) -= mutau(1);
                    }
                }
            }
        }
    }

    // body to body - wake panels
    for (auto jbody : pbl->bodies)
    {
        for (auto wake : jbody->wakes)
        {
            for (auto ew : wake->tag->elems)
            {
                // Initialisation of storage variables for the iteration
                Eigen::Matrix3d glob2loc;
                size_t nVertices = ew->nodes.size();
                Eigen::MatrixXd trsfCoordSrc(3, nVertices);

                // Computations related to source panels
                glob2loc = Glob2loc(false, ew); // Transformation matrix
                trsfCoordSrc = TrsfCoordSrc(false, ew, glob2loc); // Local coordinates of the source panel

                // wake panels
                for (auto ibody : pbl->bodies)
                {
                    for (auto ei : ibody->tag->elems)
                    {
                        // Initialisation of storage variables for the iteration
                        Eigen::Vector3d trsfCoordTgt;
                        Eigen::Vector3d distance;
                        Eigen::Vector2d wmutau;

                        // Computations related to target panels
                        trsfCoordTgt = TrsfCoordTgt(false, ei, ew, glob2loc); // Local coordinates of the target panel
                        distance = Distance(false, ei, ew, glob2loc); // Distances between source and target panels
                        wmutau = MuTau(false, ei, ew, trsfCoordSrc, trsfCoordTgt, distance); // Surface AIC

                        // Assembly of the matrices
                        A(rows.at(ei), rows.at(wake->wkMap.at(ew).first)) += wmutau(0);
                        A(rows.at(ei), rows.at(wake->wkMap.at(ew).second)) -= wmutau(0);
                    }
                }
            }
        }
    }

    // body to body - wake panels - symmetry
    if (pbl->symY)
    {
        for (auto jbody : pbl->bodies)
        {
            for (auto wake : jbody->wakes)
            {
                for (auto ew : wake->tag->elems)
                {
                    // Initialisation of storage variables for the iteration
                    Eigen::Matrix3d glob2loc;
                    size_t nVertices = ew->nodes.size();
                    Eigen::MatrixXd trsfCoordSrc(3, nVertices);

                    // Computations related to source panels
                    glob2loc = Glob2loc(true, ew); // Transformation matrix
                    trsfCoordSrc = TrsfCoordSrc(true, ew, glob2loc); // Local coordinates of the source panel

                    // wake panels
                    for (auto ibody : pbl->bodies)
                    {
                        for (auto ei : ibody->tag->elems)
                        {
                            // Initialisation of storage variables for the iteration
                            Eigen::Vector3d trsfCoordTgt;
                            Eigen::Vector3d distance;
                            Eigen::Vector2d wmutau;

                            // Computations related to target panels
                            trsfCoordTgt = TrsfCoordTgt(true, ei, ew, glob2loc); // Local coordinates of the target panel
                            distance = Distance(true, ei, ew, glob2loc); // Distances between source and target panels
                            wmutau = MuTau(true, ei, ew, trsfCoordSrc, trsfCoordTgt, distance); // Surface AIC

                            // Assembly of the matrices
                            A(rows.at(ei), rows.at(wake->wkMap.at(ew).first)) -= wmutau(0);
                            A(rows.at(ei), rows.at(wake->wkMap.at(ew).second)) += wmutau(0);
                        }
                    }
                }
            }
        }
    }

    // field to body
    if (pbl->field)
        for (auto body : pbl->bodies)
            for (auto e : body->tag->elems)
                for (size_t j = 0; j < pbl->field->tag->elems.size(); ++j)
                {
                    C(rows.at(e), j) = sigma(e, pbl->field->tag->elems[j]);
                    Cx(rows.at(e), j) = sigmaX(e, pbl->field->tag->elems[j]);
                    Cy(rows.at(e), j) = sigmaY(e, pbl->field->tag->elems[j]);
                    Cz(rows.at(e), j) = sigmaZ(e, pbl->field->tag->elems[j]);
                }

    valid = true;
}

/**
 * @brief Compute unit axes of the source panel
 */
Eigen::Matrix3d Builder::Glob2loc(bool symy, Element *ej)
{
    Eigen::Matrix3d glob2loc; // Transformation matrix
    Eigen::Vector3d n = ej->normal(); // Unit normal vector of the source panel

    // Computation of the unit longitudinal vector - Source panel
    Eigen::Vector3d cg = ej->getVMem().getCg();
    Eigen::Vector3d x0 = ej->nodes[0]->pos; // Coordinates of node 1
    Eigen::Vector3d l = (x0 - cg).normalized();

    // Computation of the unit perpendicular vector - Source panel
    Eigen::Vector3d p = (n.cross(l)).normalized();

    // Handling of the outputs
    glob2loc.row(0) = l;
    glob2loc.row(1) = p;
    glob2loc.row(2) = n;

    if(symy)
    {
        glob2loc.row(1) = -glob2loc.row(1);
        glob2loc.col(1) = -glob2loc.col(1);
    }

    return glob2loc;
}

/**
 * @brief Compute coordinates of the source panel from global to local axes
 */
Eigen::MatrixXd Builder::TrsfCoordSrc(bool symy, Element *ej, Eigen::Matrix3d const &glob2loc)
{
    size_t nVertices = ej->nodes.size();
    Eigen::MatrixXd trsfCoordSrc(3, nVertices);
    Eigen::MatrixXd coords(3, nVertices);

    for (int i = 0; i < 3; ++i)
        for (int j = 0; j < nVertices; ++j)
            coords(i, j) = ej->nodes[(nVertices - 1) - j]->pos[i]; // Global coordinates of the nodes of the source panel

    if(symy)
        coords.row(1) = -coords.row(1);

    for (int i = 0; i < nVertices; ++i)
        trsfCoordSrc.col(i) = glob2loc * coords.col(i); // Local coordinates of the nodes of the source panel

    return trsfCoordSrc;
}

/**
 * @brief Compute coordinates of the target panel from global to local axes
 */
Eigen::Vector3d Builder::TrsfCoordTgt(bool symy, Element *ei, Element *ej, Eigen::Matrix3d const &glob2loc)
{
    Eigen::Vector3d trsfCoordTgt;
    Eigen::Vector3d cg = ei->getVMem().getCg(); // Global coordinates of the CG of the target panel
    trsfCoordTgt = glob2loc * cg; // Local coordinates of the target panel

    return trsfCoordTgt;
}

/**
 * @brief Compute distances between two panels
 */
Eigen::Vector3d Builder::Distance(bool symy, Element *ei, Element *ej, Eigen::Matrix3d const &glob2loc)
{
    Eigen::Vector3d distance; // Distance between the target and source panels
    Eigen::Vector3d cgSrc = ej->getVMem().getCg();
    Eigen::Vector3d cgTgt = ei->getVMem().getCg();

    for (int i = 0; i < 3; ++i)
        distance(i) = cgTgt(i) - cgSrc(i);

    if(symy)
        distance(1) = cgTgt(1) + 2 * cgSrc(1);

    distance = glob2loc * distance;

    return distance;
}

/**
 * @brief Compute doublet AIC and source AIC from body panel ej to body panel ei
 */
Eigen::Vector2d Builder::MuTau(bool symy, Element *ei, Element *ej, Eigen::MatrixXd const &trsfCoordSrc, Eigen::Vector3d const &trsfCoordTgt, Eigen::Vector3d const &distance)
{
    double mu = 0;
    double tau = 0;
    Eigen::Vector2d mu_tau;
    Eigen::Vector4d d, e, h, r, F, G;
    size_t nVertices = ei->nodes.size();

    // Computation of the doublet AIC - mu
    if (ei->no == ej->no && !symy)
        mu = 0.5; // Doublet term - Influence of a panel on itself
    else
    {
        // Coefficients
        for (int i = 1; i <= nVertices; ++i)
        {
            e(i - 1) = (trsfCoordTgt(0) - trsfCoordSrc(0,i - 1)) * (trsfCoordTgt(0) - trsfCoordSrc(0,i - 1)) + distance(2) * distance(2);
            h(i - 1) = (trsfCoordTgt(0) - trsfCoordSrc(0,i - 1)) * (trsfCoordTgt(1) - trsfCoordSrc(1,i - 1));
            r(i - 1) = sqrt((trsfCoordTgt(0) - trsfCoordSrc(0,i - 1)) * (trsfCoordTgt(0) - trsfCoordSrc(0,i - 1)) + (trsfCoordTgt(1) - trsfCoordSrc(1,i - 1)) * (trsfCoordTgt(1) - trsfCoordSrc(1,i - 1)) + distance(2) * distance(2));
        }
        for (int i = 1; i <= nVertices; ++i)
        {
            int j = i % nVertices + 1;

            F(i - 1) = (trsfCoordSrc(1,j - 1) - trsfCoordSrc(1,i - 1)) * e(i - 1) - (trsfCoordSrc(0,j - 1) - trsfCoordSrc(0,i - 1)) * h(i - 1);
            G(i - 1) = (trsfCoordSrc(1,j - 1) - trsfCoordSrc(1,i - 1)) * e(j - 1) - (trsfCoordSrc(0,j - 1) - trsfCoordSrc(0,i - 1)) * h(j - 1);
        }
        // Doublet term - Influence of a panel on another panel
        for (int i = 1; i <= nVertices; ++i)
        {
            int j = i % nVertices + 1;

            mu = mu + atan2(distance(2) * (trsfCoordSrc(0,j - 1) - trsfCoordSrc(0,i - 1)) * (F(i - 1) * r(j - 1) - G(i - 1) * r(i - 1)),
                            distance(2) * distance(2) * (trsfCoordSrc(0,j - 1) - trsfCoordSrc(0,i - 1)) * (trsfCoordSrc(0,j - 1) - trsfCoordSrc(0,i - 1)) * r(i - 1) * r(j - 1) + F(i - 1) * G(i - 1));
        }
        mu = -1 / (4 * PI) * mu;
    }

    // Computation of the source AIC - tau
    // Coefficients
    for (int i = 1; i <= nVertices; ++i)
    {
        int j = i % nVertices + 1;

        d(i - 1) = sqrt((trsfCoordSrc(0,j - 1) - trsfCoordSrc(0,i - 1)) * (trsfCoordSrc(0,j - 1) - trsfCoordSrc(0,i - 1)) + (trsfCoordSrc(1,j - 1) - trsfCoordSrc(1,i - 1)) * (trsfCoordSrc(1,j - 1) - trsfCoordSrc(1,i - 1)));
        r(i - 1) = sqrt((trsfCoordTgt(0) - trsfCoordSrc(0,i - 1)) * (trsfCoordTgt(0) - trsfCoordSrc(0,i - 1)) + (trsfCoordTgt(1) - trsfCoordSrc(1,i - 1)) * (trsfCoordTgt(1) - trsfCoordSrc(1,i - 1)) + distance(2) * distance(2));
    }

    if (ei->no == ej->no && !symy)
    {
        // Source term - Influence of a panel on itself
        for (int i = 1; i <= nVertices; ++i)
        {
            int j = i % nVertices + 1;

            tau = tau + ((trsfCoordTgt(0) - trsfCoordSrc(0,i - 1)) * (trsfCoordSrc(1,j - 1) - trsfCoordSrc(1,i - 1)) - (trsfCoordTgt(1) - trsfCoordSrc(1,i - 1)) * (trsfCoordSrc(0,j - 1) - trsfCoordSrc(0,i - 1))) / d(i - 1) * log((r(i - 1) + r(j - 1) + d(i - 1)) / (r(i - 1) + r(j - 1) - d(i - 1)));
        }
        tau = -1 / (4 * PI) * tau;
    }
    else
    {
        // Source term - Influence of a panel on another panel
        for (int i = 1; i <= nVertices; ++i)
        {
            int j = i % nVertices + 1;

            tau = tau + ((trsfCoordTgt(0) - trsfCoordSrc(0,i - 1)) * (trsfCoordSrc(1,j - 1) - trsfCoordSrc(1,i - 1)) - (trsfCoordTgt(1) - trsfCoordSrc(1,i - 1)) * (trsfCoordSrc(0,j - 1) - trsfCoordSrc(0,i - 1))) / d(i - 1) * log((r(i - 1) + r(j - 1) + d(i - 1)) / (r(i - 1) + r(j - 1) - d(i - 1)));
        }

        tau = -1 / (4 * PI) * tau - distance(2) * mu;
    }

    mu_tau(0) = mu;
    mu_tau(1) = tau;

    return mu_tau;
}

/**
 * @brief Compute source AIC from field panel ej to body panel ei
 */
double Builder::sigma(Element *ei, Element *ej)
{
    /* Initialisation of the parameters */
    //    double sigma = 0;
    //    double A, B, C, I, II, III, R;

    /* Import of the coordinates of the influencing field cell vertices */
    //    x = 0; y = 0; z = 0;
    //    xC = 0; yC = 0; zC = 0;

    /* Field source potential influence between two field panels */
    //    for (int i = 1; i <= 2; ++i) {
    //        for (int j = 1; j <= 2; ++j) {
    //            for (int k = 1; k <= 2; ++k) {
    //                A = x - xC[i-1];
    //                B = y - yC[j-1];
    //                C = z - zC[k-1];
    //                R = sqrt(A * A + B * B + C * C);
    //                I = B * C * log(A + R) - A * A / 2 * atan((B * C) / (A * R));
    //                II = C * A * log(B + R) - B * B / 2 * atan((C * A) / (B * R));
    //                III = A * B * log(C + R) - C * C / 2 * atan((A * B) / (C * R));

    //                sigma -= pow((-1),(double)(i + j + k)) * (I + II + III);
    //            }
    //        }
    //    }

    //    sigma = - 1 / (4 * PI) * sigma;

    //    return sigma;

    return 0;
}

/**
 * @brief Compute source x-velocity AIC from field panel ej to body panel ei
 */
double Builder::sigmaX(Element *ei, Element *ej)
{
    /* Initialisation of the parameters */
    //    double sigmaX = 0;
    //    double A, B, C, R, t0;

    /* Import of the coordinates of the influencing field cell vertices */
    //    x = 0; y = 0; z = 0;
    //    xC = 0; yC = 0; zC = 0;

    /* Field source velocity influence between a field panel and a body panel */
    //    for (int i = 1; i <= 2; ++i) {
    //        for (int j = 1; j <= 2; ++j) {
    //            for (int k = 1; k <= 2; ++k) {
    //                A = (x - xC[i-1]);
    //                B = (y - yC[j-1]);
    //                C = (z - zC[k-1]);
    //                R = sqrt(A * A + B * B + C * C);
    //                t0 = B / 2 * log((R + C) / (R - C)) + C / 2 * log((R + B) / (R - B)) - A * atan((B * C) / (A * R));

    //                sigmaX += pow((-1),(double)(i + j + k)) * t0;
    //            }
    //        }
    //    }

    //    sigmaX = 1 / (4 * PI) * sigmaX;

    //    return sigmaX;

    return 0;
}

/**
 * @brief Compute source y-velocity AIC from field panel ej to body panel ei
 */
double Builder::sigmaY(Element *ei, Element *ej)
{
    /* Initialisation of the parameters */
    //    double sigmaY = 0;
    //    double A, B, C, R, t1;

    /* Import of the coordinates of the influencing field cell vertices */
    //    x = 0; y = 0; z = 0;
    //    xC = 0; yC = 0; zC = 0;

    /* Field source velocity influence between a field panel and a body panel */
    //    for (int i = 1; i <= 2; ++i) {
    //        for (int j = 1; j <= 2; ++j) {
    //            for (int k = 1; k <= 2; ++k) {
    //                A = (x - xC[i-1]);
    //                B = (y - yC[j-1]);
    //                C = (z - zC[k-1]);
    //                R = sqrt(A * A + B * B + C * C);
    //                t1 = C / 2 * log((R + A) / (R - A)) + A / 2 * log((R + C) / (R - C)) - B * atan((C * A) / (B * R));

    //                sigmaY += pow((-1),(double)(i + j + k)) * t1;
    //            }
    //        }
    //    }

    //    sigmaY = 1 / (4 * PI) * sigmaY;

    //    return sigmaY;

    return 0;
}

/**
 * @brief Compute source z-velocity AIC from field panel ej to body panel ei
 */
double Builder::sigmaZ(Element *ei, Element *ej)
{
    /* Initialisation of the parameters */
    //    double sigmaZ = 0;
    //    double A, B, C, R, t2;

    /* Import of the coordinates of the influencing field cell vertices */
    //    x = 0; y = 0; z = 0;
    //    xC = 0; yC = 0; zC = 0;

    /* Field source velocity influence between a field panel and a body panel */
    //    for (int i = 1; i <= 2; ++i) {
    //        for (int j = 1; j <= 2; ++j) {
    //            for (int k = 1; k <= 2; ++k) {
    //                A = (x - xC[i-1]);
    //                B = (y - yC[j-1]);
    //                C = (z - zC[k-1]);
    //                R = sqrt(A * A + B * B + C * C);
    //                t2 = A / 2 * log((R + B) / (R - B)) + B / 2 * log((R + A) / (R - A)) - C * atan((A * B) / (C * R));

    //                sigmaZ += pow((-1),(double)(i + j + k)) * t2;
    //            }
    //        }
    //    }

    //    sigmaZ = 1 / (4 * PI) * sigmaZ;

    //    return sigmaZ;

    return 0;
}

void Builder::write(std::ostream &out) const
{
    out << "fpm::Builder with AIC matrices:"
        << "\n\tA(" << A.rows() << "," << A.cols() << "), B(" << B.rows() << "," << B.cols() << ")"
        << "\n\tC(" << C.rows() << "," << C.cols() << ")"
        << "\n\tCx(" << Cx.rows() << "," << Cx.cols() << "), Cy(" << Cy.rows() << "," << Cy.cols() << "), Cz(" << Cz.rows() << "," << Cz.cols() << ")"
        << "\nwith\t" << *pbl;
}
