/*
 * Copyright 2020 University of Liège
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "fWake.h"
#include "fEdge.h"
#include "wNode.h"
#include "wElement.h"
#include <vector>
#include <unordered_set>
using namespace tbox;
using namespace fpm;

Wake::Wake(std::shared_ptr<MshData> _msh, std::string const &name, Tag *const &wTag) : Group(_msh, name)
{
    // Create set of TE edges
    std::unordered_set<Edge *, EquEdge, EquEdge> teEdges;
    for (auto e : tag->elems)
    {
        Edge *ed = new Edge({e->nodes[0], e->nodes[3]});
        ed->el0 = e;
        teEdges.insert(ed);
    }

    // Find wing elements having an edge on the TE
    for (auto e : wTag->elems)
    {
        size_t idx = 0;
        for (size_t i = 0; i < e->nodes.size(); ++i)
        {
            // build wing edge
            std::vector<Node *> eN(2);
            for (size_t j = 0; j < 2; ++j)
                eN[j] = e->nodes[(idx + j) % e->nodes.size()];
            Edge ed(eN);
            // check if wing edge is a TE edge
            auto it = teEdges.find(&ed);
            if (it != teEdges.end())
            {
                // check normal direction
                if (e->normal().dot((*it)->el0->normal()) > 0)
                    (*it)->el1 = e;
                else
                    (*it)->el2 = e;
            }
            idx++;
        }
    }

    // Create a map between the wing and wake elements
    for (auto ed : teEdges)
    {
        if (ed->el1 == NULL || ed->el2 == NULL)
        {
            std::stringstream err;
            err << "fpm::Wake: no wing element from tag " << *wTag << "could be associated to wake element " << *(ed->el0) << "!\n";
            throw std::runtime_error(err.str());
        }
        wkMap[ed->el0] = std::pair<Element *, Element *>(ed->el1, ed->el2);
        delete ed;
    }
}

void Wake::write(std::ostream &out) const
{
    out << "fpm::Wake " << *tag << std::endl;
}
